/* FastCounter 1.0
An example of fast animation using a backing storage bitmap,
and a separate thread for timing (used to use Pulse, leaving
behind some confusingly named functions, no longer used for
the standard Pulse message), and alpha blending for a nice
fade-out effect.  For fun, try resizing the window while it
is running.  By AGMS, July 2000.  Public domain and not
called AGMSFastCounter since it's just a hack. */

#include <Application.h>
#include <Bitmap.h>
#include <View.h>
#include <Window.h>
#include <stdio.h>

static thread_id ViewPulseThreadID = -1;
/* The ID of the thread which periodically calls the CounterView's
   Pulse function to do the animation and invalidate the view (but
   not do any drawing or CPU intensive activity).  Negative if
   it doesn't exist. */

static volatile bool TimeToDie = false;
/* When set to TRUE, the threads will try to exit soon. */



class CounterApp : public BApplication
{
public:
  CounterApp () : BApplication ("application/x-vnd.agmsmith.Counter") {};
  virtual void ReadyToRun();
};


class CounterWindow : public BWindow 
{
public:
  CounterWindow () : BWindow (BRect (30, 30, 600, 400), "Counter",
  B_TITLED_WINDOW, 0)
  {
  };
  ~CounterWindow ()
  {
    status_t ReturnCode;
    
    TimeToDie = true;
    if (ViewPulseThreadID > 0)
      wait_for_thread (ViewPulseThreadID, &ReturnCode);
  };
  virtual bool QuitRequested ()
  {
    be_app->PostMessage (B_QUIT_REQUESTED);
    return true;
  };
};


int32 ViewPulserCode (void *PassedInData); // Forward reference.


class CounterView : public BView
{
public:
  CounterView (BRect FrameSize) : BView (FrameSize, "CounterView", B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
    m_MovingDotPoint (10, 20),
    m_BackingView (FrameSize, "BackingView", B_FOLLOW_ALL_SIDES, B_FULL_UPDATE_ON_RESIZE)
  {
    m_CurrentCount = 0;
    m_StringResizeNeededCount = 10;
    m_BackingBitmap = NULL;
    SetViewColor (B_TRANSPARENT_COLOR);
    FrameResized (FrameSize.right, FrameSize.bottom); // Set m_TextStartPoint.
    ViewPulseThreadID = spawn_thread (ViewPulserCode,
      "CounterViewPulser", 80 /* Priority */, this); // Creates but doesn't start.
  };
  ~CounterView ()
  {
    if (m_BackingBitmap != NULL)
    {
      m_BackingBitmap->RemoveChild (&m_BackingView);
      delete m_BackingBitmap;
      m_BackingBitmap = NULL;
    }
  };
  virtual void Draw (BRect updateRect);
  virtual void FrameResized (float width, float height);
  virtual void Pulse (void);
private:
  int m_CurrentCount;
  int m_StringResizeNeededCount;
  BRect m_BndRect;
  BPoint m_TextStartPoint;
  BPoint m_MovingDotPoint;
  int m_MovingDotSize;
  float m_MoveSpeed;
  BBitmap *m_BackingBitmap;
  BView m_BackingView;
};


/* This is the function which the view pulsing thread executes. */

int32 ViewPulserCode (void *PassedInData)
{
  int          i;
  bigtime_t    CurrentTime;
  bigtime_t    WakeupTime = 0;
  BWindow     *WindowOfView;
  CounterView *ViewPntr;

  ViewPntr = (CounterView *) PassedInData;
  WindowOfView = ViewPntr->Window ();
  for (i = 0; i < 20000 && !TimeToDie; i++)
  {
    WakeupTime += 10000;
    WindowOfView->Lock ();
    ViewPntr->Pulse ();
    WindowOfView->Unlock ();
    
    // Go to sleep until the next update time, accounting for
    // time wasted locking the window.
    
    CurrentTime = system_time ();
    if (WakeupTime < CurrentTime)
      WakeupTime = CurrentTime;
    else
      snooze_until (WakeupTime, B_SYSTEM_TIMEBASE);
  }
  return B_OK; // Return code of thread.
}


int main (int argc, char** argv)
{
  CounterApp theApp;
  theApp.Run();
  return 0;
}


void CounterApp::ReadyToRun ()
{
  CounterWindow *MyWin = new CounterWindow;
  MyWin->AddChild (new CounterView (MyWin->Bounds ()));
  MyWin->Show ();
  if (ViewPulseThreadID > 0)
    resume_thread (ViewPulseThreadID);
}


void CounterView::Draw (BRect updateRect)
{
  BRect MovingRect (
    m_MovingDotPoint.x,
    m_MovingDotPoint.y,
    m_MovingDotPoint.x + m_MovingDotSize,
    m_MovingDotPoint.y + m_MovingDotSize);
  char TempString [40];

  if (m_BackingBitmap != NULL)
  {
    m_BackingBitmap->Lock ();
    m_BackingView.SetHighColor (60, 60, 255, 8);
    m_BackingView.FillRect (m_BndRect);
    m_BackingView.SetHighColor (255, 255, 0, 255);
    m_BackingView.MovePenTo (m_TextStartPoint);
    sprintf (TempString, "%d", m_CurrentCount);
    m_BackingView.DrawString (TempString);
    m_BackingView.FillRect (MovingRect);
    m_BackingView.Sync ();
    m_BackingBitmap->Unlock ();
    MovePenTo (0, 0);
    DrawBitmap (m_BackingBitmap);
  }
}


void CounterView::Pulse (void)
{
  m_CurrentCount++;
  if (m_CurrentCount >= m_StringResizeNeededCount)
  {
    m_StringResizeNeededCount *= 10;
    FrameResized (m_BndRect.right - m_BndRect.left,
      m_BndRect.bottom - m_BndRect.top);
  }

  // Move the dot around the edge of the window, counterclockwise.
  
  if (m_MovingDotPoint.x  <= m_BndRect.left) // Moving along left side.
  {
    m_MovingDotPoint.x  = m_BndRect.left;
    if (m_MovingDotPoint.y + m_MovingDotSize >= m_BndRect.bottom)
    {
      m_MovingDotPoint.y = m_BndRect.bottom - m_MovingDotSize;
      m_MovingDotPoint.x = m_BndRect.left + m_MoveSpeed;
    }
    else
      m_MovingDotPoint.y += m_MoveSpeed;
  }
  else if (m_MovingDotPoint.y <= m_BndRect.top) // Moving along top side.
  {
    m_MovingDotPoint.y = m_BndRect.top;
    if (m_MovingDotPoint.x <= m_BndRect.left)
    {
      m_MovingDotPoint.y = m_BndRect.top + m_MoveSpeed;
      m_MovingDotPoint.x = m_BndRect.left;
    }
    else
      m_MovingDotPoint.x -= m_MoveSpeed;
  }
  else if (m_MovingDotPoint.y + m_MovingDotSize >= m_BndRect.bottom) // Moving along bottom side.
  {
    m_MovingDotPoint.y = m_BndRect.bottom - m_MovingDotSize;
    if (m_MovingDotPoint.x + m_MovingDotSize >= m_BndRect.right)
    {
      m_MovingDotPoint.y = m_BndRect.bottom - m_MovingDotSize - m_MoveSpeed;
      m_MovingDotPoint.x = m_BndRect.right - m_MovingDotSize;
    }
    else
      m_MovingDotPoint.x += m_MoveSpeed;
  }
  else // Moving along right side.
  {
    m_MovingDotPoint.x = m_BndRect.right - m_MovingDotSize;
    if (m_MovingDotPoint.y <= m_BndRect.top)
    {
      m_MovingDotPoint.y = m_BndRect.top;
      m_MovingDotPoint.x = m_BndRect.right - m_MovingDotSize - m_MoveSpeed;
    }
    else
      m_MovingDotPoint.y -= m_MoveSpeed;
  }

  Invalidate ();
}


void CounterView::FrameResized (float width, float height)
{
  BRect BitmapRect (0, 0, width, height);
  char  TempString [40];

  m_BndRect = Bounds ();

  m_MovingDotSize = (int) (height / 20);
  if (m_MovingDotSize < 1)
    m_MovingDotSize = 1;
  m_MoveSpeed = m_MovingDotSize / 2.0;
  
  // Resize the offscreen bitmap and its view.

  if (m_BackingBitmap != NULL)
  {
    m_BackingBitmap->RemoveChild (&m_BackingView);
    delete m_BackingBitmap;
    m_BackingBitmap = NULL;
  }

  m_BackingView.ResizeTo (width, height);

  m_BackingBitmap = new BBitmap (BitmapRect, B_RGBA32, true /* Accepts subviews */);
  if (!m_BackingBitmap->IsValid ())
  {
    delete m_BackingBitmap;
    m_BackingBitmap = NULL;
  }
  else
  {
    m_BackingBitmap->AddChild (&m_BackingView);
    m_BackingBitmap->Lock ();
    m_BackingView.SetDrawingMode (B_OP_ALPHA);
    m_BackingView.SetFontSize (height * 0.8);
    sprintf (TempString, "%d", m_CurrentCount);
    m_TextStartPoint.x = width / 2 - m_BackingView.StringWidth (TempString) / 2;
    m_TextStartPoint.y = height / 2 + height * 0.25;
    m_BackingBitmap->Unlock ();
  }
}
